/*
 * Copyright 2018 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.firebase.auth.hash;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import com.google.common.collect.ImmutableList;
import com.google.firebase.auth.hash.Argon2.Argon2HashType;
import java.util.List;
import org.junit.Test;

public class InvalidHashTest {

  private static final byte[] SIGNER_KEY = "key".getBytes();
  private static final byte[] SALT_SEPARATOR = "separator".getBytes();

  @Test
  public void testInvalidHmac() {
    List<Hmac.Builder> builders = ImmutableList.<Hmac.Builder>of(
        HmacSha512.builder(),
        HmacSha256.builder(),
        HmacSha1.builder(),
        HmacMd5.builder()
    );
    for (Hmac.Builder builder : builders) {
      try {
        builder.build();
        fail("No error thrown for missing key");
      } catch (Exception expected) {
        assertTrue(expected instanceof IllegalArgumentException);
      }
    }
  }

  @Test
  public void testInvalidRepeatableHash() {
    // TODO(rsgowman): Once we can update to Java8, we could just do something like this instead of
    // having all of the helpers:
    //     assertThrows(IllegalArgumentException.class, ()-> Md5.builder().setRounds(-1).build());

    List<RepeatableHash.Builder> builders = ImmutableList.<RepeatableHash.Builder>builder()
        .add(Sha512.builder().setRounds(0))
        .add(Sha256.builder().setRounds(0))
        .add(Sha1.builder().setRounds(0))
        .add(Md5.builder().setRounds(-1))
        .add(Pbkdf2Sha256.builder().setRounds(-1))
        .add(PbkdfSha1.builder().setRounds(-1))
        .add(Sha512.builder().setRounds(8193))
        .add(Sha256.builder().setRounds(8193))
        .add(Sha1.builder().setRounds(8193))
        .add(Md5.builder().setRounds(8193))
        .add(Pbkdf2Sha256.builder().setRounds(120001))
        .add(PbkdfSha1.builder().setRounds(120001))
        .build();
    for (RepeatableHash.Builder builder : builders) {
      try {
        builder.build();
        fail("No error thrown for invalid rounds");
      } catch (Exception expected) {
        assertTrue(expected instanceof IllegalArgumentException);
      }
    }
  }

  @Test
  public void testValidRepeatableHash() {
    Md5.builder().setRounds(0).build();
    Md5.builder().setRounds(8192).build();
    Sha1.builder().setRounds(1).build();
    Sha1.builder().setRounds(8192).build();
    Sha256.builder().setRounds(1).build();
    Sha256.builder().setRounds(8192).build();
    Sha512.builder().setRounds(1).build();
    Sha512.builder().setRounds(8192).build();
    PbkdfSha1.builder().setRounds(0).build();
    PbkdfSha1.builder().setRounds(120000).build();
    Pbkdf2Sha256.builder().setRounds(0).build();
    Pbkdf2Sha256.builder().setRounds(120000).build();
  }

  @Test
  public void testInvalidScrypt() {
    List<Scrypt.Builder> builders = ImmutableList.of(
        Scrypt.builder() // missing signer key
            .setSaltSeparator(SALT_SEPARATOR)
            .setRounds(5)
            .setMemoryCost(12),
        Scrypt.builder() // invalid rounds (> 8)
            .setKey(SIGNER_KEY)
            .setSaltSeparator(SALT_SEPARATOR)
            .setRounds(9)
            .setMemoryCost(14),
        Scrypt.builder() // invalid rounds (< 0)
            .setKey(SIGNER_KEY)
            .setSaltSeparator(SALT_SEPARATOR)
            .setRounds(-1)
            .setMemoryCost(14),
        Scrypt.builder() // invalid memory cost (> 15)
            .setKey(SIGNER_KEY)
            .setSaltSeparator(SALT_SEPARATOR)
            .setRounds(8)
            .setMemoryCost(15)
    );
    for (Scrypt.Builder builder : builders) {
      try {
        builder.build();
        fail("No error thrown for invalid configuration");
      } catch (IllegalArgumentException expected) {
        // expected
      }
    }
  }

  @Test
  public void testInvalidArgon2() {
    List<Argon2.Builder> builders = ImmutableList.of(
        Argon2.builder() // hashLengthBytes < 4
            .setHashLengthBytes(2)
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(8)
            .setIterations(16)
            .setMemoryCostKib(512),
        Argon2.builder() // hashLengthBytes > 1024
            .setHashLengthBytes(2048)
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(8)
            .setIterations(16)
            .setMemoryCostKib(512),
        Argon2.builder() // missing hashType
            .setHashLengthBytes(32)
            .setParallelism(8)
            .setIterations(16)
            .setMemoryCostKib(512),
        Argon2.builder() // parallelism < 1
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(0)
            .setHashLengthBytes(32)
            .setIterations(16)
            .setMemoryCostKib(512),
        Argon2.builder() // parallelism > 16
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(32)
            .setHashLengthBytes(32)
            .setIterations(16)
            .setMemoryCostKib(512),
        Argon2.builder() // iterations < 1
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(16)
            .setHashLengthBytes(32)
            .setIterations(0)
            .setMemoryCostKib(512),
        Argon2.builder() // iterations > 16
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(16)
            .setHashLengthBytes(32)
            .setIterations(32)
            .setMemoryCostKib(512),
        Argon2.builder() // memoryCostKib < 1
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(16)
            .setHashLengthBytes(32)
            .setIterations(8)
            .setMemoryCostKib(0),
        Argon2.builder() // memoryCostKib > 32768
            .setHashType(Argon2HashType.ARGON2_ID)
            .setParallelism(16)
            .setHashLengthBytes(32)
            .setIterations(8)
            .setMemoryCostKib(32769)
    );
    for (Argon2.Builder builder : builders) {
      try {
        builder.build();
        fail("No error thrown for invalid configuration");
      } catch (IllegalArgumentException expected) {
        // expected
      }
    }
  }
}
